/* viterbi.c -- viterbi maximum likelihood detector.
 * (C) 2000, C. Bond. All rights reserved.
 *
 * Assumes nominal sample values are: +2.0, 0.0, -2.0.
 *
 */


#include <stdio.h>

double dp0,p00,p01,p10,p11;	/* state variables */
int up[256];			/* upper survivor path */
int lp[256];			/* lower survivor path */
int pl;				/* survivor path length */

/* Viterbi decoder. Internal variables are global so that external
 * analysis is possible. If the input sample vector size, 'n' is
 * unity, it is assumed that step mode is desired and no initialization
 * of globals takes place. Hence, the calling routine must initialize
 * all the above variables to zero. (Only the first element of the
 * arrays needs to be initialized.)
 * If 'n' is greater than unity, it is assumed that batch decoding
 * is desired and the global variables are then initialized.
 */
int viterbi(double *samples,int *bits,int n)
{
	double r;
	int j,k,i;
	
/* check for valid parameters */
	if (n < 1) return 1;

/* initialize variables, unless n = 1 */
	if (n > 1) {
		up[0] = 0;
		lp[0] = 0;
		pl = 0;
		dp0 = p00 = p01 = 0.0;
	}
	for (i = 0;i < n;i ++) {
		r = samples[i];
		if (p00 < (p10 + r + 1.0))
			p01 = p00;
		else
			p01 = p10 + r + 1.0;
		if ((p00 - r + 1.0) < p10)
			p11 = p00 - r + 1.0;
		else
			p11 = p10;
		if (1.0 <= (dp0 - r)) {
			k = 3;
			dp0 = r + 1.0;
		}
		else if ((-1.0 < (dp0 - r)) && ((dp0 - r) < 1.0)) {
			k = 2;
		}
		else if ((dp0 - r) <= -1.0) {
			k = 1;
			dp0 = r - 1.0;
		}
		p00 = p01;
		p10 = p11;
		switch (k) {
			case 1:
				for (j = 0;j <= pl; j++)
					bits[i-pl+j]=lp[j];
				up[0] = 1;
				lp[0] = 0;
				pl = 1;
				break;
			case 2:
				up[pl] = 0;
				lp[pl++] = 0;
				break;
			case 3:
				for (j = 0;j <= pl; j++)
					bits[i-pl+j] = up[j];
				up[0] = 0;
				lp[0] = 1;
				pl = 1;
				break;
		}
	}
	return 0;
}
